"""Non-terminal symbols of Python grammar (from "graminit.h")."""

from typing import Final

single_input: Final[int]
file_input: Final[int]
eval_input: Final[int]
decorator: Final[int]
decorators: Final[int]
decorated: Final[int]
async_funcdef: Final[int]
funcdef: Final[int]
parameters: Final[int]
typedargslist: Final[int]
tfpdef: Final[int]
varargslist: Final[int]
vfpdef: Final[int]
stmt: Final[int]
simple_stmt: Final[int]
small_stmt: Final[int]
expr_stmt: Final[int]
annassign: Final[int]
testlist_star_expr: Final[int]
augassign: Final[int]
del_stmt: Final[int]
pass_stmt: Final[int]
flow_stmt: Final[int]
break_stmt: Final[int]
continue_stmt: Final[int]
return_stmt: Final[int]
yield_stmt: Final[int]
raise_stmt: Final[int]
import_stmt: Final[int]
import_name: Final[int]
import_from: Final[int]
import_as_name: Final[int]
dotted_as_name: Final[int]
import_as_names: Final[int]
dotted_as_names: Final[int]
dotted_name: Final[int]
global_stmt: Final[int]
nonlocal_stmt: Final[int]
assert_stmt: Final[int]
compound_stmt: Final[int]
async_stmt: Final[int]
if_stmt: Final[int]
while_stmt: Final[int]
for_stmt: Final[int]
try_stmt: Final[int]
with_stmt: Final[int]
with_item: Final[int]
except_clause: Final[int]
suite: Final[int]
test: Final[int]
test_nocond: Final[int]
lambdef: Final[int]
lambdef_nocond: Final[int]
or_test: Final[int]
and_test: Final[int]
not_test: Final[int]
comparison: Final[int]
comp_op: Final[int]
star_expr: Final[int]
expr: Final[int]
xor_expr: Final[int]
and_expr: Final[int]
shift_expr: Final[int]
arith_expr: Final[int]
term: Final[int]
factor: Final[int]
power: Final[int]
atom_expr: Final[int]
atom: Final[int]
testlist_comp: Final[int]
trailer: Final[int]
subscriptlist: Final[int]
subscript: Final[int]
sliceop: Final[int]
exprlist: Final[int]
testlist: Final[int]
dictorsetmaker: Final[int]
classdef: Final[int]
arglist: Final[int]
argument: Final[int]
comp_iter: Final[int]
comp_for: Final[int]
comp_if: Final[int]
encoding_decl: Final[int]
yield_expr: Final[int]
yield_arg: Final[int]
sync_comp_for: Final[int]
func_body_suite: Final[int]
func_type: Final[int]
func_type_input: Final[int]
namedexpr_test: Final[int]
typelist: Final[int]
sym_name: Final[dict[int, str]]
